package at.ac.uibk.gitsearch.web.rest;

import static org.junit.Assert.assertTrue;
import static org.springframework.security.test.web.servlet.request.SecurityMockMvcRequestPostProcessors.csrf;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;
import static org.springframework.test.web.servlet.result.MockMvcResultHandlers.print;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import at.ac.uibk.gitsearch.GitsearchApp;
import at.ac.uibk.gitsearch.repository.search.testESService.ElasticSearchTestConfiguration;
import at.ac.uibk.gitsearch.security.AuthoritiesConstants;
import at.ac.uibk.gitsearch.service.PluginManagementService;
import at.ac.uibk.gitsearch.service.SearchService;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketInfoDTO;
import at.ac.uibk.gitsearch.service.ShoppingBasketService.ShoppingBasketRedirectInfoDTO;
import at.ac.uibk.gitsearch.web.rest.PluginInterfaceResource.SearchResultsDTOMapper;
import java.io.IOException;
import java.util.concurrent.TimeUnit;
import org.codeability.sharing.plugins.api.SharingPluginConfig;
import org.codeability.sharing.plugins.api.UserPrincipal;
import org.codeability.sharing.plugins.api.search.SearchInputDTO;
import org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO;
import org.codeability.sharing.plugins.api.search.SearchRequestDTO;
import org.codeability.sharing.plugins.api.search.SearchResultDTO;
import org.codeability.sharing.plugins.api.search.SearchResultsDTO;
import org.elasticsearch.node.NodeValidationException;
import org.junit.Assert;
import org.junit.jupiter.api.AfterAll;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.api.Timeout;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.http.MediaType;
import org.springframework.security.test.context.support.WithMockUser;
import org.springframework.test.web.servlet.MockMvc;

/**
 * Integration tests for the {@link PluginInterfaceResource} REST controller.
 */
@AutoConfigureMockMvc
@SpringBootTest(classes = GitsearchApp.class)
public class PluginInterfaceResourceIT {

    @Autowired
    private PluginManagementService pluginManagementService;

    @Autowired
    ElasticSearchTestConfiguration elasticSearchTestConfiguration;

    static ElasticSearchTestConfiguration staticElasticSearchTestConfiguration;

    @BeforeEach // must be started as BeforeEach, in order to autowire the
    // ElasticSearchTestConfiguration
    @Timeout(value = 2, unit = TimeUnit.MINUTES)
    public void setUpESServer() throws IOException, NodeValidationException {
        if (staticElasticSearchTestConfiguration == null) {
            elasticSearchTestConfiguration.startTestNode();
            staticElasticSearchTestConfiguration = elasticSearchTestConfiguration;
        }
    }

    // setup Plugin info
    private static SharingPluginConfig testPluginInfo = new SharingPluginConfig(
        "testPlugin",
        new SharingPluginConfig.Action[] {
            new SharingPluginConfig.Action("testActionId", "http://localhost/actionTransfer", "testAction", "unused FilterExpression"),
        }
    );

    @AfterAll
    public static void shutDownESServer() {
        if (staticElasticSearchTestConfiguration != null) {
            staticElasticSearchTestConfiguration.stopTestContainer();
            staticElasticSearchTestConfiguration = null;
        }
    }

    @Value("${testing.testUser.name}")
    private String userTestName;

    @Value("${testing.testUser.gitLabToken}")
    private String userTestToken;

    @Autowired
    private MockMvc restPluginInterfaceMockMvc;

    @Autowired
    private SearchService searchService;

    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER)
    public void testGetRedirectURL() throws Exception {
        // just load everything from index :-)
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO searchResultPage = searchService.searchResultPage(searchQuery);
        Assert.assertTrue("At least one hit?", searchResultPage.getSearchResult().size() >= 1);

        ShoppingBasketInfoDTO sbi = new ShoppingBasketInfoDTO(searchResultPage.getSearchResult().toArray(new SearchResultDTO[] {}));
        sbi.setPlugin(testPluginInfo.pluginName);
        sbi.setAction(testPluginInfo.actions[0].actionId);

        restPluginInterfaceMockMvc
            .perform(
                post("/api//pluginIF/getPluginRedirectInfos")
                    .content(TestUtil.convertObjectToJsonBytes(sbi))
                    .with(csrf().asHeader())
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andDo(print())
            .andExpect(status().is2xxSuccessful())
            .andExpect(
                TestUtil.testResult(
                    ShoppingBasketRedirectInfoDTO.class,
                    result -> {
                        String redirectURL = result.getRedirectURL();
                        assertTrue(redirectURL.contains("returnURL="));
                        assertTrue(redirectURL.contains("apiBaseURL="));
                        // further checks would be nice
                    }
                )
            );
    }

    @Test
    public void testSearchPageDetails() throws Exception {
        // searches for all :-)

        final org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO searchMetadata = new org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO(
            null,
            null,
            null,
            null,
            null,
            null
        );
        final org.codeability.sharing.plugins.api.search.SearchInputDTO searchQuery = new org.codeability.sharing.plugins.api.search.SearchInputDTO(
            null,
            searchMetadata,
            null,
            null,
            null,
            0
        );

        // This test may fail, if token is expired!
        final SearchRequestDTO searchRequest = new SearchRequestDTO(new UserPrincipal(userTestName, userTestToken), searchQuery);

        restPluginInterfaceMockMvc
            .perform(
                post("/api/pluginIF/v0.1/page-details")
                    .content(TestUtil.convertObjectToJsonBytes(searchRequest))
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andDo(print())
            .andExpect(status().is2xxSuccessful())
            .andExpect(
                TestUtil.testResult(
                    org.codeability.sharing.plugins.api.search.SearchResultsDTO.class,
                    result -> {
                        // not really a good test :-(
                        assertTrue("At least three results", result.getSearchResult().size() >= 3);
                    }
                )
            );
    }

    @Test
    public void testSearchPageDetailsWithInvalidToken() throws Exception {
        // searches for all :-)

        final org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO searchMetadata = new org.codeability.sharing.plugins.api.search.SearchInputMetadataDTO(
            null,
            null,
            null,
            null,
            null,
            null
        );
        final org.codeability.sharing.plugins.api.search.SearchInputDTO searchQuery = new org.codeability.sharing.plugins.api.search.SearchInputDTO(
            null,
            searchMetadata,
            null,
            null,
            null,
            0
        );

        // This test may fail, if token is expired!
        final SearchRequestDTO searchRequest = new SearchRequestDTO(
            new UserPrincipal("Michael Breu (with invalid Testtoken)", "XXXXX"),
            searchQuery
        );

        restPluginInterfaceMockMvc
            .perform(
                post("/api/pluginIF/v0.1/page-details")
                    .content(TestUtil.convertObjectToJsonBytes(searchRequest))
                    .contentType(MediaType.APPLICATION_JSON)
            )
            .andDo(print())
            .andExpect(status().is2xxSuccessful())
            .andExpect(
                TestUtil.testResult(
                    org.codeability.sharing.plugins.api.search.SearchResultsDTO.class,
                    result -> {
                        // not really a good test :-(
                        assertTrue("Currently some public result", result.getSearchResult().size() >= 3);
                        // result.getSearchResult().forEach(r -> r.getProject()..getNamespace());
                    }
                )
            );
    }

    /**
     * just a simple test of the results mapper
     */
    @Test
    @WithMockUser(authorities = AuthoritiesConstants.USER)
    public void testSearchResultsMapper() throws IOException {
        final SearchResultsDTOMapper resultsMapper = SearchResultsDTOMapper.INSTANCE;

        // just load everything from index :-)
        final SearchInputMetadataDTO searchMetadata = new SearchInputMetadataDTO(null, null, null, null, null, null);
        SearchInputDTO searchQuery = new SearchInputDTO(null, searchMetadata, null, null, null, 0);
        SearchResultsDTO results = searchService.searchResultPage(searchQuery);
        Assert.assertTrue("At least one hit?", results.getSearchResult().size() >= 1);

        org.codeability.sharing.plugins.api.search.SearchResultsDTO mappedResults = resultsMapper.searchResultsToPlugin(results);

        // punktuelle Tests
        Assert.assertEquals(results.getPageStartIndex(), mappedResults.getPageStartIndex());
        Assert.assertEquals(results.getHitCount(), mappedResults.getHitCount());
        Assert.assertEquals(results.getSearchResult().size(), mappedResults.getSearchResult().size());
        Assert.assertEquals(
            results.getSearchResult().get(0).getFile().getIndexing_date(),
            mappedResults.getSearchResult().get(0).getFile().getIndexing_date()
        );
    }
}
